
--------------- Hud function implementations ---------------

local function DrawMag(player, x, y, width, height, empty, bac)

    --Slug Repeater 
    --Draw background
    v.drawFill(slug_x, slug_y, slug_width, slug_height, background_mag_color)
    if(stplayer.slug_mag > 0) then
        v.drawFill(slug_x, slug_y, slug_loadedbased_width, slug_height, loaded_mag_color)
    else
        v.drawFill(slug_x, slug_y, slug_emptybased_width, slug_height, empty_mag_color)
    end
end

local function drawweaponhud(v, stplayer, cam)
    
    if(not Valid(stplayer.mo, "jazz") or stplayer.tf_mode ~= TF_MODES.BOT_MODE) then
       return false 
    end

    local norings_mag_color = 72
    local loaded_mag_color = 135
    local empty_mag_color = 41
    local background_mag_color = 31

    
    -- Adjust to render the bar just below the RINGS HUD
    local slug_x = hudinfo[HUD_RINGS].x
    local slug_y = hudinfo[HUD_RINGS].y + 20
    local slug_height = 3
    local slug_width = 50
    --The ratio that make the bar "move"
    local slug_mag_ratio = FixedDiv(stplayer.slug_mag*FRACUNIT, SLUG_MAG_MAX*FRACUNIT)
    local slug_reload_ratio = FixedDiv(stplayer.slug_reload_tics*FRACUNIT, SLUG_RELOAD_TICS_MAX*FRACUNIT)
    local slug_ring_ratio = FixedDiv(stplayer.rings*FRACUNIT, SLUG_RING_REQUIREMENT*FRACUNIT)


    --The adjusted width
    local slug_loadedbased_width = FixedMul(slug_width, slug_mag_ratio)
    local slug_emptybased_width = FixedMul(slug_width, slug_reload_ratio)
    local slug_ring_width = FixedMul(slug_width, slug_ring_ratio)

    --Slug Repeater
    v.drawFill(slug_x, slug_y, slug_width+1, slug_height+1, background_mag_color)
    if(stplayer.rings < SLUG_RING_REQUIREMENT) then
        v.drawFill(slug_x, slug_y, slug_ring_width, slug_height, norings_mag_color)
    elseif(stplayer.slug_mag > 0) then
        v.drawFill(slug_x, slug_y, slug_loadedbased_width, slug_height, loaded_mag_color)
    else
        v.drawFill(slug_x, slug_y, slug_emptybased_width, slug_height, empty_mag_color)
    end
    
    
    -- Adjust to render the bar just below the RINGS HUD
    local crescent_x = slug_x
    local crescent_y = slug_y + slug_height*2
    local crescent_height = slug_height
    local crescent_width = slug_width
    --The ratio that make the bar "move"
    local crescent_mag_ratio = FixedDiv(stplayer.crescent_mag*FRACUNIT, CRESCENT_MAG_MAX*FRACUNIT)
    local crescent_reload_ratio = FixedDiv(stplayer.crescent_reload_tics*FRACUNIT, CRESCENT_RELOAD_TICS_MAX*FRACUNIT)
    local crescent_ring_ratio = FixedDiv(stplayer.rings*FRACUNIT, CRESCENT_RING_REQUIREMENT*FRACUNIT)

    --The adjusted width
    local crescent_loadedbased_width = FixedMul(crescent_width, crescent_mag_ratio)
    local crescent_emptybased_width = FixedMul(crescent_width, crescent_reload_ratio)
    local crescent_ring_width = FixedMul(crescent_width, crescent_ring_ratio)

    --Crescent Repeater 
    v.drawFill(crescent_x, crescent_y, crescent_width+1, crescent_height+1, background_mag_color)
    if(stplayer.rings < CRESCENT_RING_REQUIREMENT) then
        v.drawFill(crescent_x, crescent_y, crescent_ring_width, crescent_height, norings_mag_color)
    elseif(stplayer.crescent_mag > 0) then
        v.drawFill(crescent_x, crescent_y, crescent_loadedbased_width, crescent_height, loaded_mag_color)
    else
        v.drawFill(crescent_x, crescent_y, crescent_emptybased_width, crescent_height, empty_mag_color)
    end


end


--------------- Hud function definitions ---------------


hud.add(drawweaponhud, "game")


local function SpawnEmptyMag(player)
    
    local icon = P_SpawnMobjFromMobj(player.mo, 0, 0, player.mo.height, MT_EMPT)

    -- print(player.empty_icon)
    --If icon already exists
    if(icon ~= nil and icon.target ~= nil and icon.target.valid == true) then
        P_KillMobj(icon.target)
    end

    icon.target = player.mo
    
end

--Empties the mag
local function EmptyMag(player)
    player.slug_mag = 0 --Empty mag
    player.crescent_mag = 0 --Empty mag
    player.slug_reload_tics = SLUG_RELOAD_TICS_MAX
    player.crescent_reload_tics = CRESCENT_RELOAD_TICS_MAX 
end

---------------------------------------------------------------------
---------- Connecting all of the hooks from this file  --------------
---------------------------------------------------------------------

--Makes the player shoot a single bullet projectile
local function ShootBullet(player)

    player.slug_mag = $-1 --Detect a bullet being shot

    -- Bullet to be shot
    local bullet = P_SpawnMobjFromMobj(player.mo, 0, 0, 30*FRACUNIT, MT_WBUL)
    bullet.scale = player.mo.scale
    bullet.target = player.mo
    bullet.angle = player.mo.angle

    --The total speed that should be distributed to horizontal and vertical speed depending on vertical angle
    local total_speed = mobjinfo[bullet.type].speed
    local side_spread = (player.speed/FRACUNIT)/8 -- Furthest side spread depending on speed
    
    --Forward and vertical speeds
    local bullet_forward_speed  = FixedMul(total_speed, cos(player.aiming))
    local bullet_vertical_speed = FixedMul(total_speed, sin(player.aiming))
    local bullet_side_speed = P_RandomRange(-side_spread, side_spread)*FRACUNIT

    --Shooting the bullet bullet is under water 
    if(bullet.eflags & MFE_UNDERWATER ~= 0) then
        bullet_side_speed = $*2
        bullet_forward_speed = $/2
    end

    -- P_SetObjectMomZ(bullet, bullet_vertical_speed, false)
    P_InstaThrust(bullet, player.mo.angle + ANGLE_90, bullet_side_speed) -- Bullet spread horizontally
    P_SetObjectMomZ(bullet, bullet_vertical_speed, bullet_vertical_speed) -- Side spead
    P_Thrust(bullet, player.mo.angle, bullet_forward_speed) -- Forward movement speed

    --Play souund
    S_StartSound(player.mo, sfx_bullet) -- Bullet sound
end



--Makes the players shoot the heavy blast projectile
local function ShootBlast(player)

    player.crescent_mag = $-1 --Detect a blast being shot
    
    -- blast to be shot
    local blast = P_SpawnMobjFromMobj(player.mo, 0, 0, 30*FRACUNIT, MT_WBLS)
    blast.scale = player.mo.scale
    blast.target = player.mo
    blast.angle = player.mo.angle

    --The total speed that should be distributed to horizontal and vertical speed depending on vertical angle
    local total_speed = mobjinfo[blast.type].speed
    local side_spread = (player.speed/FRACUNIT)/8 -- Furthest side spread depending on speed
    
    --Forward and vertical speeds
    local blast_forward_speed  = FixedMul(total_speed, cos(player.aiming))
    local blast_vertical_speed = FixedMul(total_speed, sin(player.aiming))
    local blast_side_speed = P_RandomRange(-side_spread, side_spread)*FRACUNIT

    --Shooting the blast blast is under water 
    if(blast.eflags & MFE_UNDERWATER ~= 0) then
        blast_side_speed = $*2
        blast_forward_speed = $/2
    end

    -- P_SetObjectMomZ(blast, blast_vertical_speed, false)
    P_InstaThrust(blast, player.mo.angle + ANGLE_90, blast_side_speed) -- blast spread horizontally
    P_SetObjectMomZ(blast, blast_vertical_speed, blast_vertical_speed) -- Side spead
    P_Thrust(blast, player.mo.angle, blast_forward_speed) -- Forward movement speed

    --Play souund
    S_StartSound(player.mo, sfx_blast) -- blast sound
end


--Thinker for when the player is in the BulletStance state
local function WeaponStanceThink(player)
    if(player.mo.state == S_WEAPON_STANCE) then
        
        --Makes sure to display correct sprite depending on whether Jazz is airbone or not
        if(not P_IsObjectOnGround(player.mo)) then
            player.mo.sprite2 = SPR2_BUJU
        elseif(P_IsObjectOnGround(player.mo)) then
            player.mo.sprite2 = SPR2_BUSN
            
            --Skid on the ground
            if(player.speed > 0) then
                SpawnOnSide(player, MT_DUST)
            end
        end



        --Hold long is required to hold the shoot button to switch from Slug to Crescent
        local weapon_swap_tic = states[player.mo.state].tics/3

        --Trigger primary and/or secondary weapon
        if(player.spinheld == 0 and player.prevspinheld ~= 0 and player.prevspinheld < weapon_swap_tic) then
            player.burst_shot_interval_tics = BURST_SHOT_INTERVAL_TIC_MAX -- It will trigger shooting bullets since we have several to shoot
        elseif(player.prevspinheld == weapon_swap_tic) then
            --Do we have ammo?
            if(player.crescent_mag > 0 and player.rings >= CRESCENT_RING_REQUIREMENT) then
                ShootBlast(player) -- Single shot
            else
                --The Crescent Mag is EMPTY!!!!!!!
                S_StartSound(player.mo, sfx_empty) 
                SpawnEmptyMag(player)
            end
        end

        --Shoots a single bullet at the time
        if(player.burst_shot_interval_tics ~= 0 and player.burst_shot_interval_tics % 3 == 0) then

            --Check ammo
            if(player.slug_mag > 0 and player.rings >= SLUG_RING_REQUIREMENT) then 
                ShootBullet(player)
            else
                --No ammo, reset
                player.burst_shot_interval_tics = 0
                S_StartSound(player.mo, sfx_empty)
                SpawnEmptyMag(player)
            end
        end
        
        --Counting down the bullet timer 
        if(player.burst_shot_interval_tics > 0) then
            player.burst_shot_interval_tics = $-1
        end
    end    
end


local function BotPlayerThink(player)

    
    --Landing effect
    if(player.mo.eflags & MFE_JUSTHITFLOOR and not S_SoundPlaying(player.mo, sfx_land)) then
        S_StartSound(player.mo, sfx_land)
        SpawnOnSide(player, MT_SPINDUST)
    end
    
    --If taking damage, then emptying the mag
    if(player.mo.state == mobjinfo[player.mo.type].painstate) then
       EmptyMag(player)
    end

    if(Active(player)) then

        --Start shooting the Bullet projectile
        if(player.spinheld == 1) then
            player.mo.state = S_WEAPON_STANCE
        end
        --Execute the weapon stance thinker
        WeaponStanceThink(player)

        

        --How Jazz is affected in robot mode while under special boost 
        if(player.special_boost_tics > 0) then --Special speed boost is in the affect
            player.normalspeed = skins[player.mo.skin].normalspeed + SPECIAL_BOOST_BOT_SPEED
            player.acceleration = skins[player.mo.skin].acceleration + SPECIAL_BOOST_BOT_ACCEL
            P_SpawnGhostMobj(player.mo)
        else --Special speed boost is removed
            player.normalspeed = skins[player.mo.skin].normalspeed
            player.acceleration = skins[player.mo.skin].acceleration
        end

        
        --If not yet climbing, either start climbing or stop climbing
        if(player.climbing == 0) then
            
            if(player.lastlinehit ~= -1 and player.powers[pw_pushing] ~= 0 --If hitting a wall
            and player.tf_mode == TF_MODES.BOT_MODE --While in the robot mode
            and player.cmd.buttons & BT_JUMP ~= 0 --While holding jump button
            and player.jumpspecial >= 1 --If special is ready
            and (abs(player.cmd.forwardmove) > 20 or abs(player.cmd.sidemove) > 20)) then --While moving significantly
                
                
                --Information about the wall
                local climb_wall = lines[player.lastlinehit]
                --Extra checks for the wall
                if(climb_wall.flags & ML_NOCLIMB == 0 and climb_wall.special ~= 41) then
                
                    --Wall x and y points
                    local wall_x, wall_y = P_ClosestPointOnLine(player.mo.x, player.mo.y, climb_wall)
                    --Angle between Jazz position and the closes point on the wall
                    local wall_angle = R_PointToAngle2(player.mo.x, player.mo.y, wall_x, wall_y)
                    --Distance between Jazz and closest point on the walll
                    local wall_distance = R_PointToDist2(player.mo.x, player.mo.y, wall_x, wall_y)

                    player.charability = CA_GLIDEANDCLIMB
                    player.climbing = 2
                    player.jumpspecial = 0


                    S_StartSound(player.mo, sfx_climb)
                end
            
            --Stop climbing by removing the ability
            elseif(player.charability == CA_GLIDEANDCLIMB) then
                player.charability = skins[player.mo.skin].ability
                -- player.jumpspecial = 1
            end
        end

    end
end


local function BlastMobjThink(blast)
    
    local ghost = P_SpawnGhostMobj(blast)
    
    -- local scaling = FRACUNIT
    if(blast.state == mobjinfo[blast.type].deathstate) then
        
        ghost.scale = blast.scale-2*FRACUNIT
        
    end
    
end

local function BlastMoveCollide(blast, collide_mo)
    
end

local function BulletMobjThinker(bullet)
    --Underwater spread of the bullet
    if(bullet.eflags & MFE_UNDERWATER) then
        bullet.momx = 9*$/10
        bullet.momy = 9*$/10
    end
    P_SpawnGhostMobj(bullet)
end

local function BulletMoveBlocked(bullet, enemy, wall)
    -- Bullet flies into the wall
    if(wall ~= nil) then 
        S_StartSound(wall.frontsector, sfx_miss)    
    -- Bullet hits the entity
    else
        S_StartSound(enemy, sfx_hit)
    end
end

addHook("MobjThinker", function (iconmo)
    -- if(iconmo.valid == true and iconmo.player ~= nil and iconmo.player.valid == true) then
    if(iconmo.target ~= nil and iconmo.target.valid == true) then
        P_MoveOrigin(iconmo, iconmo.target.x, iconmo.target.y, iconmo.target.z + iconmo.target.height)
    else
        P_KillMobj(iconmo)
    end
end, MT_EMPT)

addHook("PlayerThink", function(player) 
    if(Valid(player.mo, "jazz") and player.tf_mode == TF_MODES.BOT_MODE) then
        BotPlayerThink(player) 
    end
end)

addHook("MobjThinker", function(mo)
    BulletMobjThinker(mo)
end, MT_WBUL)

addHook("MobjMoveBlocked", function(mo, tmmo, line)
    BulletMoveBlocked(mo, tmmo, line)
end, MT_WBUL)

addHook("MobjThinker", function(mo)
    BlastMobjThink(mo)
end, MT_WBLS)

addHook("MobjMoveCollide", function(mo, collide_mo)
    BlastMoveCollide(mo, collide_mo)
end, MT_WBLS)