---------------------------------------------------------------------
---------------- Functions and hook definitions  --------------------
---------------------------------------------------------------------


local function TransformationPlayerThink(player)

    --Count down the transformation tics to 0
    if(player.transformation_tics > 0) then
        player.transformation_tics = $-1
    end

    

    if(Active(player)) then
        local transform_input = nil 

        if(player == secondarydisplayplayer) then
            transform_input = input.gameControl2Down(GC_CUSTOM1) or input.gameControl2Down(GC_CUSTOM2)
        else
            transform_input = input.gameControlDown(GC_CUSTOM1) or input.gameControlDown(GC_CUSTOM2)
        end

        --Stop breakdancing if not holding transform or movement buttoms
        if(player.mo.state == S_TF_TO_BOT_BR and player.mo.tics <= states[player.mo.state].tics/2 and player.spinheld == 0) then
            player.mo.state = states[player.mo.state].nextstate
        end

        -- Try to transform if not already transforming 
        if(player.mo.state ~= S_TF_TO_BOT and player.mo.state ~= S_TF_TO_CAR and player.transformation_tics <= 0
        and player.mo.state ~= mobjinfo[player.mo.type].painstate) then

            -- Transformation from the car mode to the bot mode
            if(player.tf_mode == TF_MODES.CAR_MODE) then
                if(P_IsObjectOnGround(player.mo) and player.spinheld == 2) then
                    player.mo.state = S_TF_TO_BOT_BR
                elseif(player.jumpheld == 2 or transform_input) then
                    player.mo.state = S_TF_TO_BOT
                end
            -- Transformation from the bot mode to the car mode
            elseif(player.tf_mode == TF_MODES.BOT_MODE and player.powers[pw_carry] == 0 and player.climbing == 0
            and (transform_input or (player.jumpheld == 3 and player.jumpspecial > 0 and not P_IsObjectOnGround(player.mo)))) then
                player.mo.state = S_TF_TO_CAR

            end 
        end
    end
    -- Execute while in the transformation state
    if(player.mo.state == S_TF_TO_BOT or player.mo.state == S_TF_TO_CAR) then
        player.pflags = $|PF_JUMPSTASIS
        player.jumpspecial = 0
    end    
    
end

local function TransformationPreThink(player)

    --While breakdancig, keep keep the momentum and guard up
    if(player.mo.state == S_TF_TO_BOT_BR) then
        player.powers[pw_strong] = $|STR_GUARD
        player.mo.friction = FRACUNIT-FRACUNIT/50
        player.cmd.sidemove    = clamp($, -10, 10)
        player.cmd.forwardmove = clamp($, -10, 10)
        -- P_InstaThrust(player.mo, R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy), 30*FRACUNIT)
    else
        player.powers[pw_strong] = $&~STR_GUARD
    end

end


local function TransformationPostThink(player)

    --Manually stopping the transforming animation
    if((player.mo.state == S_TF_TO_CAR or player.mo.state == S_TF_TO_BOT) and player.mo.tics < states[player.mo.state].tics/2 and player.mo.frame & FF_FRAMEMASK == A) then
        player.mo.state = states[player.mo.state].nextstate
    end



    --Assign the correct state after transformation
    if(player.mo.state == S_TF_TO_BOT) then
        if(not P_IsObjectOnGround(player.mo)) then
            states[player.mo.state].nextstate = S_PLAY_JUMP
        --Keep walking 
        elseif(player.mo.player.speed > FRACUNIT) then -- At low speed on the ground
            --Play walking animation
            states[player.mo.state].nextstate = S_PLAY_WALK
        --Transforming at standstill
        else
            --Play normal animations
            states[player.mo.state].nextstate = S_PLAY_STND
        end
    end
end


local function TransformMoveCollide(jazzmo, collide_mo)
    
    --Hit the enemy while spinning and on guard
    if(L_ZCollide(jazzmo, collide_mo) and jazzmo.state == S_TF_TO_BOT_BR and jazzmo.player.powers[pw_strong] & STR_GUARD) then
        P_DamageMobj(collide_mo, jazzmo, jazzmo, 1)
    end

end


---------------------------------------------------------------------
---------- Connecting all of the hooks from this file  --------------
---------------------------------------------------------------------


addHook("PlayerThink", function(player) 
    if(not Valid(player.mo, "jazz")) then
        return 
    end
    TransformationPlayerThink(player) 
end)

addHook("PreThinkFrame", function(player) 
    for player in players.iterate() do
        if(not Valid(player.mo, "jazz")) then
            return 
        end

        TransformationPreThink(player) 
    end
end)

addHook("PostThinkFrame", function(player) 
    for player in players.iterate() do
        if(not Valid(player.mo, "jazz")) then
            return 
        end

        TransformationPostThink(player) 
    end
end)


addHook("MobjMoveCollide", function(mo, collide_mo)
    if(not Valid(mo, "jazz")) then
        return --Jazz is not the one that is colliding
    end

    TransformMoveCollide(mo, collide_mo)

end, MT_PLAYER)

